@@ -6,10 +6,10 @@ Pyro - Python Remote Objects.  Copyright by Irmen de Jong (irmen@razorvine.net).
 
 from __future__ import print_function, division
 import inspect
-import collections
 import re
 import logging
 import sys
+import ssl
 import os
 import time
 import threading
@@ -19,12 +19,10 @@ import warnings
 import socket
 import random
 from Pyro4 import errors, socketutil, util, constants, message, futures
-from Pyro4.socketserver.threadpoolserver import SocketServer_Threadpool
-from Pyro4.socketserver.multiplexserver import SocketServer_Multiplex
 from Pyro4.configuration import config
 
 
-__all__ = ["URI", "Proxy", "Daemon", "current_context", "callback", "batch", "async", "expose", "behavior",
+__all__ = ["URI", "Proxy", "Daemon", "current_context", "callback", "batch", "asyncproxy", "expose", "behavior",
            "oneway", "SerializedBlob", "_resolve", "_locateNS"]
 
 if sys.version_info >= (3, 0):
@@ -145,7 +143,8 @@ class URI(object):
     def __eq__(self, other):
         if not isinstance(other, URI):
             return False
-        return (self.protocol, self.object, self.sockname, self.host, self.port) == (other.protocol, other.object, other.sockname, other.host, other.port)
+        return (self.protocol, self.object, self.sockname, self.host, self.port) ==\
+               (other.protocol, other.object, other.sockname, other.host, other.port)
 
     def __ne__(self, other):
         return not self.__eq__(other)
@@ -223,7 +222,9 @@ class Proxy(object):
          "_pyroRawWireResponse", "_pyroHandshake", "_pyroMaxRetries", "_pyroSerializer", "_Proxy__async",
          "_Proxy__pyroHmacKey", "_Proxy__pyroTimeout", "_Proxy__pyroConnLock"])
 
-    def __init__(self, uri):
+    def __init__(self, uri, connected_socket=None):
+        if connected_socket:
+            uri = URI("PYRO:" + uri + "@<<connected-socket>>:0")
         if isinstance(uri, basestring):
             uri = URI(uri)
         elif not isinstance(uri, URI):
@@ -236,7 +237,7 @@ class Proxy(object):
         self._pyroOneway = set()  # oneway-methods of the remote object, gotten from meta-data
         self._pyroSeq = 0  # message sequence number
         self._pyroRawWireResponse = False  # internal switch to enable wire level responses
-        self._pyroHandshake = "hello"  # the data object that should be sent in the initial connection handshake message (can be any serializable object)
+        self._pyroHandshake = "hello"  # the data object that should be sent in the initial connection handshake message
         self._pyroMaxRetries = config.MAX_RETRIES
         self.__pyroHmacKey = None
         self.__pyroTimeout = config.COMMTIMEOUT
@@ -245,6 +246,8 @@ class Proxy(object):
         self.__async = False
         current_context.annotations = {}
         current_context.response_annotations = {}
+        if connected_socket:
+            self.__pyroCreateConnection(False, connected_socket)
 
     @property
     def _pyroHmacKey(self):
@@ -296,7 +299,7 @@ class Proxy(object):
 
     def __repr__(self):
         if self._pyroConnection:
-            connected = "connected "+self._pyroConnection.family()
+            connected = "connected " + self._pyroConnection.family()
         else:
             connected = "not connected"
         return "<%s.%s at 0x%x; %s; for %s>" % (self.__class__.__module__, self.__class__.__name__,
@@ -308,7 +311,7 @@ class Proxy(object):
     def __getstate_for_dict__(self):
         encodedHmac = None
         if self._pyroHmacKey is not None:
-            encodedHmac = "b64:"+(base64.b64encode(self._pyroHmacKey).decode("ascii"))
+            encodedHmac = "b64:" + (base64.b64encode(self._pyroHmacKey).decode("ascii"))
         # for backwards compatibility reasons we also put the timeout and maxretries into the state
         return self._pyroUri.asString(), tuple(self._pyroOneway), tuple(self._pyroMethods), tuple(self._pyroAttrs),\
             self.__pyroTimeout, encodedHmac, self._pyroHandshake, self._pyroMaxRetries, self._pyroSerializer
@@ -332,8 +335,8 @@ class Proxy(object):
 
     def __getstate__(self):
         # for backwards compatibility reasons we also put the timeout and maxretries into the state
-        return self._pyroUri, self._pyroOneway, self._pyroMethods, self._pyroAttrs, self.__pyroTimeout,\
-               self._pyroHmacKey, self._pyroHandshake, self._pyroMaxRetries, self._pyroSerializer
+        return self._pyroUri, self._pyroOneway, self._pyroMethods, self._pyroAttrs, self.__pyroTimeout, \
+            self._pyroHmacKey, self._pyroHandshake, self._pyroMaxRetries, self._pyroSerializer
 
     def __setstate__(self, state):
         # Note that the timeout and maxretries are also part of the state (for backwards compatibility reasons),
@@ -390,6 +393,8 @@ class Proxy(object):
         """release the connection to the pyro daemon"""
         with self.__pyroConnLock:
             if self._pyroConnection is not None:
+                if self._pyroConnection.keep_open:
+                    return
                 self._pyroConnection.close()
                 self._pyroConnection = None
                 log.debug("connection released")
@@ -435,7 +440,8 @@ class Proxy(object):
             if methodname in self._pyroOneway:
                 flags |= message.FLAGS_ONEWAY
             self._pyroSeq = (self._pyroSeq + 1) & 0xffff
-            msg = message.Message(message.MSG_INVOKE, data, serializer.serializer_id, flags, self._pyroSeq, annotations=annotations, hmac_key=self._pyroHmacKey)
+            msg = message.Message(message.MSG_INVOKE, data, serializer.serializer_id, flags, self._pyroSeq,
+                                  annotations=annotations, hmac_key=self._pyroHmacKey)
             if config.LOGWIRE:
                 _log_wiredata(log, "proxy wiredata sending", msg)
             try:
@@ -467,7 +473,7 @@ class Proxy(object):
                     if msg.flags & message.FLAGS_EXCEPTION:
                         if sys.platform == "cli":
                             util.fixIronPythonExceptionForPickle(data, False)
-                        raise data
+                        raise data  # if you see this in your traceback, you should probably inspect the remote traceback as well
                     else:
                         return data
             except (errors.CommunicationError, KeyboardInterrupt):
@@ -486,23 +492,27 @@ class Proxy(object):
             log.error(err)
             raise errors.ProtocolError(err)
 
-    def __pyroCreateConnection(self, replaceUri=False):
+    def __pyroCreateConnection(self, replaceUri=False, connected_socket=None):
         """
         Connects this proxy to the remote Pyro daemon. Does connection handshake.
         Returns true if a new connection was made, false if an existing one was already present.
         """
-        with self.__pyroConnLock:
-            if self._pyroConnection is not None:
-                return False  # already connected
-            uri = _resolve(self._pyroUri, self._pyroHmacKey)
-            # socket connection (normal or Unix domain socket)
-            conn = None
-            log.debug("connecting to %s", uri)
-            connect_location = uri.sockname or (uri.host, uri.port)
+        def connect_and_handshake(conn):
             try:
                 if self._pyroConnection is not None:
                     return False  # already connected
-                sock = socketutil.createSocket(connect=connect_location, reuseaddr=config.SOCK_REUSE, timeout=self.__pyroTimeout, nodelay=config.SOCK_NODELAY)
+                if config.SSL:
+                    sslContext = socketutil.getSSLcontext(clientcert=config.SSL_CLIENTCERT,
+                                                          clientkey=config.SSL_CLIENTKEY,
+                                                          keypassword=config.SSL_CLIENTKEYPASSWD,
+                                                          cacerts=config.SSL_CACERTS)
+                else:
+                    sslContext = None
+                sock = socketutil.createSocket(connect=connect_location,
+                                               reuseaddr=config.SOCK_REUSE,
+                                               timeout=self.__pyroTimeout,
+                                               nodelay=config.SOCK_NODELAY,
+                                               sslContext=sslContext)
                 conn = socketutil.SocketConnection(sock, uri.object)
                 # Do handshake.
                 serializer = util.get_serializer(self._pyroSerializer or config.SERIALIZER)
@@ -558,7 +568,7 @@ class Proxy(object):
                     if replaceUri:
                         self._pyroUri = uri
                     self._pyroValidateHandshake(handshake_response)
-                    log.debug("connected to %s - %s", self._pyroUri, conn.family())
+                    log.debug("connected to %s - %s - %s", self._pyroUri, conn.family(), "SSL" if sslContext else "unencrypted")
                     if msg.annotations:
                         self._pyroResponseAnnotations(msg.annotations, msg.type)
                 else:
@@ -566,6 +576,24 @@ class Proxy(object):
                     err = "cannot connect to %s: invalid msg type %d received" % (connect_location, msg.type)
                     log.error(err)
                     raise errors.ProtocolError(err)
+
+        with self.__pyroConnLock:
+            if self._pyroConnection is not None:
+                return False  # already connected
+            if connected_socket:
+                if config.SSL and not isinstance(connected_socket, ssl.SSLSocket):
+                    raise socket.error("SSL configured for Pyro but existing socket is not a SSL socket")
+                uri = self._pyroUri
+            else:
+                uri = _resolve(self._pyroUri, self._pyroHmacKey)
+            # socket connection (normal or Unix domain socket)
+            conn = None
+            log.debug("connecting to %s", uri)
+            connect_location = uri.sockname or (uri.host, uri.port)
+            if connected_socket:
+                self._pyroConnection = socketutil.SocketConnection(connected_socket, uri.object, True)
+            else:
+                connect_and_handshake(conn)
             if config.METADATA:
                 # obtain metadata if this feature is enabled, and the metadata is not known yet
                 if self._pyroMethods or self._pyroAttrs:
@@ -605,7 +633,8 @@ class Proxy(object):
         self._pyroMethods = set(metadata["methods"])
         self._pyroAttrs = set(metadata["attrs"])
         if log.isEnabledFor(logging.DEBUG):
-            log.debug("from meta: methods=%s, oneway methods=%s, attributes=%s", sorted(self._pyroMethods), sorted(self._pyroOneway), sorted(self._pyroAttrs))
+            log.debug("from meta: methods=%s, oneway methods=%s, attributes=%s",
+                      sorted(self._pyroMethods), sorted(self._pyroOneway), sorted(self._pyroAttrs))
         if not self._pyroMethods and not self._pyroAttrs:
             raise errors.PyroError("remote object doesn't expose any methods or attributes. Did you forget setting @expose on them?")
 
@@ -632,10 +661,22 @@ class Proxy(object):
         """returns a helper class that lets you create batched method calls on the proxy"""
         return _BatchProxyAdapter(self)
 
-    def _pyroAsync(self, async=True):
-        """turns the proxy into async mode so you can do asynchronous method calls"""
-        self.__async = async
-        return self
+    def _pyroAsync(self, asynchronous=True):
+        """turns the proxy into asynchronous mode so you can do asynchronous method calls,
+        or sets it back to normal sync mode if you set asynchronous=False."""
+        self.__async = asynchronous
+
+    if sys.version_info < (3, 7):
+        # async keyword backwards compatibility
+        _pyroAsync_37 = _pyroAsync
+
+        def _pyroAsync(self, asynchronous=True, **kwargs):
+            if kwargs:
+                kword = list(kwargs.keys())
+                if kword != ["async"]:
+                    raise TypeError("_pyroAsync() got an unexpected keyword argument '{:s}'".format(kword[0]))
+                asynchronous = kwargs["async"]
+            return Proxy._pyroAsync_37(self, asynchronous)
 
     def _pyroInvokeBatch(self, calls, oneway=False):
         flags = message.FLAGS_BATCH
@@ -747,7 +788,8 @@ class _StreamResultIterator(object):
         if self.proxy and self.proxy._pyroConnection is not None:
             if self.pyroseq == self.proxy._pyroSeq:
                 # we're still in sync, it's okay to use the same proxy to close this stream
-                self.proxy._pyroInvoke("close_stream", [self.streamId], {}, flags=message.FLAGS_ONEWAY, objectId=constants.DAEMON_NAME)
+                self.proxy._pyroInvoke("close_stream", [self.streamId], {},
+                                       flags=message.FLAGS_ONEWAY, objectId=constants.DAEMON_NAME)
             else:
                 # The proxy's sequence number has diverged.
                 # One of the reasons this can happen is because this call is being done from python's GC where
@@ -756,7 +798,8 @@ class _StreamResultIterator(object):
                 # We create a temporary second proxy to call close_stream on. This is inefficient, but avoids the problem.
                 try:
                     with self.proxy.__copy__() as closingProxy:
-                        closingProxy._pyroInvoke("close_stream", [self.streamId], {}, flags=message.FLAGS_ONEWAY, objectId=constants.DAEMON_NAME)
+                        closingProxy._pyroInvoke("close_stream", [self.streamId], {},
+                                                 flags=message.FLAGS_ONEWAY, objectId=constants.DAEMON_NAME)
                 except errors.CommunicationError:
                     pass
         self.proxy = None
@@ -808,10 +851,10 @@ class _BatchProxyAdapter(object):
             else:
                 yield result  # it is a regular result object, yield that and continue.
 
-    def __call__(self, oneway=False, async=False):
-        if oneway and async:
+    def __call__(self, oneway=False, asynchronous=False):
+        if oneway and asynchronous:
             raise errors.PyroError("async oneway calls make no sense")
-        if async:
+        if asynchronous:
             return _AsyncRemoteMethod(self, "<asyncbatch>", self.__proxy._pyroMaxRetries)()
         else:
             results = self.__proxy._pyroInvokeBatch(self.__calls, oneway)
@@ -819,6 +862,20 @@ class _BatchProxyAdapter(object):
             if not oneway:
                 return self.__resultsgenerator(results)
 
+    if sys.version_info < (3, 7):
+        # async keyword backwards compatibility
+        call_37 = __call__
+
+        def __call__(self, oneway=False, **kwargs):
+            if kwargs:
+                kword = list(kwargs.keys())
+                if kword != ["async"] and kword != ["asynchronous"]:
+                    raise TypeError("__call__() got an unexpected keyword argument '{:s}'".format(kword[0]))
+                if kword == ["async"]:
+                    kwargs = {"asynchronous": kwargs["async"]}
+            kwargs["oneway"] = oneway
+            return _BatchProxyAdapter.call_37(self, **kwargs)
+
     def _pyroInvoke(self, name, args, kwargs):
         # ignore all parameters, we just need to execute the batch
         results = self.__proxy._pyroInvokeBatch(self.__calls)
@@ -827,7 +884,7 @@ class _BatchProxyAdapter(object):
 
 
 class _AsyncRemoteMethod(object):
-    """async method call abstraction (call will run in a background thread)"""
+    """asynchronous method call abstraction (call will run in a background thread)"""
     def __init__(self, proxy, name, max_retries):
         self.__proxy = proxy
         self.__name = name
@@ -849,7 +906,7 @@ class _AsyncRemoteMethod(object):
                 # use a copy of the proxy otherwise calls would still be done in sequence,
                 # and use contextmanager to close the proxy after we're done
                 with self.__proxy.__copy__() as proxy:
-                    delay = 0.1 + random.random()/5
+                    delay = 0.1 + random.random() / 5
                     while not proxy._pyroConnection:
                         try:
                             proxy._pyroBind()
@@ -857,8 +914,8 @@ class _AsyncRemoteMethod(object):
                             if "no free workers" not in str(x):
                                 raise
                             time.sleep(delay)   # wait a bit until a worker might be available again
-                            delay += 0.4 + random.random()/2
-                            if 0 < config.COMMTIMEOUT/2 < delay:
+                            delay += 0.4 + random.random() / 2
+                            if 0 < config.COMMTIMEOUT / 2 < delay:
                                 raise
                     value = proxy._pyroInvoke(self.__name, args, kwargs)
                 asyncresult.value = value
@@ -866,11 +923,11 @@ class _AsyncRemoteMethod(object):
             except (errors.ConnectionClosedError, errors.TimeoutError) as x:
                 # only retry for recoverable network errors
                 if attempt >= self.__max_retries:
-                    # ignore any exceptions here, return them as part of the async result instead
+                    # ignore any exceptions here, return them as part of the asynchronous result instead
                     asyncresult.value = futures._ExceptionWrapper(x)
                     return
             except Exception as x:
-                # ignore any exceptions here, return them as part of the async result instead
+                # ignore any exceptions here, return them as part of the asynchronous result instead
                 asyncresult.value = futures._ExceptionWrapper(x)
                 return
 
@@ -880,9 +937,9 @@ def batch(proxy):
     return proxy._pyroBatch()
 
 
-def async(proxy, async=True):
-    """convenience method to set proxy to async mode"""
-    return proxy._pyroAsync(async)
+def asyncproxy(proxy, asynchronous=True):
+    """convenience method to set proxy to asynchronous or sync mode."""
+    proxy._pyroAsync(asynchronous)
 
 
 def pyroObjectToAutoProxy(obj):
@@ -927,7 +984,16 @@ def expose(method_or_class):
             raise AttributeError("exposing private names (starting with _) is not allowed")
         func._pyroExposed = True
         return method_or_class
-    if util.is_private_attribute(method_or_class.__name__):
+    attrname = getattr(method_or_class, "__name__", None)
+    if not attrname:
+        # we could be dealing with a descriptor (classmethod/staticmethod), this means the order of the decorators is wrong
+        if inspect.ismethoddescriptor(method_or_class):
+            attrname = method_or_class.__get__(None, dict).__name__
+            raise AttributeError("using @expose on a classmethod/staticmethod must be done "
+                                 "after @classmethod/@taticmethod. Method: " + attrname)
+        else:
+            raise AttributeError("@expose cannot determine what this is: " + repr(method_or_class))
+    if util.is_private_attribute(attrname):
         raise AttributeError("exposing private names (starting with _) is not allowed")
     if inspect.isclass(method_or_class):
         clazz = method_or_class
@@ -961,7 +1027,7 @@ def behavior(instance_mode="session", instance_creator=None):
         if not inspect.isclass(clazz):
             raise TypeError("behavior decorator can only be used on a class")
         if instance_mode not in ("single", "session", "percall"):
-            raise ValueError("invalid instance mode: "+instance_mode)
+            raise ValueError("invalid instance mode: " + instance_mode)
         if instance_creator and not callable(instance_creator):
             raise TypeError("instance_creator must be a callable")
         clazz._pyroInstancing = (instance_mode, instance_creator)
@@ -1006,7 +1072,9 @@ class DaemonObject(object):
                 # Something seems wrong: nothing is remotely exposed.
                 # Possibly because older code not using @expose is now running with a more recent Pyro version
                 # where @expose is mandatory in the default configuration. Give a hint to the user.
-                warnings.warn("Class %r doesn't expose any methods or attributes. Did you forget setting @expose on them?" % type(obj))
+                if not inspect.isclass(obj):
+                    obj = type(obj)
+                warnings.warn("Class %r doesn't expose any methods or attributes. Did you forget setting @expose on them?" % obj)
             return metadata
         else:
             log.debug("unknown object requested: %s", objectId)
@@ -1037,28 +1105,38 @@ class Daemon(object):
     to the appropriate objects.
     """
 
-    def __init__(self, host=None, port=0, unixsocket=None, nathost=None, natport=None, interface=DaemonObject):
-        if host is None:
-            host = config.HOST
-        if nathost is None:
-            nathost = config.NATHOST
-        if natport is None:
-            natport = config.NATPORT or None
-        if nathost and unixsocket:
-            raise ValueError("cannot use nathost together with unixsocket")
-        if (nathost is None) ^ (natport is None):
-            raise ValueError("must provide natport with nathost")
-        if config.SERVERTYPE == "thread":
-            self.transportServer = SocketServer_Threadpool()
-        elif config.SERVERTYPE == "multiplex":
-            self.transportServer = SocketServer_Multiplex()
+    def __init__(self, host=None, port=0, unixsocket=None, nathost=None, natport=None, interface=DaemonObject, connected_socket=None):
+        if connected_socket:
+            nathost = natport = None
         else:
-            raise errors.PyroError("invalid server type '%s'" % config.SERVERTYPE)
+            if host is None:
+                host = config.HOST
+            if nathost is None:
+                nathost = config.NATHOST
+            if natport is None:
+                natport = config.NATPORT or None
+            if nathost and unixsocket:
+                raise ValueError("cannot use nathost together with unixsocket")
+            if (nathost is None) ^ (natport is None):
+                raise ValueError("must provide natport with nathost")
         self.__mustshutdown = threading.Event()
         self.__mustshutdown.set()
         self.__loopstopped = threading.Event()
         self.__loopstopped.set()
-        self.transportServer.init(self, host, port, unixsocket)
+        if connected_socket:
+            from Pyro4.socketserver.existingconnectionserver import SocketServer_ExistingConnection
+            self.transportServer = SocketServer_ExistingConnection()
+            self.transportServer.init(self, connected_socket)
+        else:
+            if config.SERVERTYPE == "thread":
+                from Pyro4.socketserver.threadpoolserver import SocketServer_Threadpool
+                self.transportServer = SocketServer_Threadpool()
+            elif config.SERVERTYPE == "multiplex":
+                from Pyro4.socketserver.multiplexserver import SocketServer_Multiplex
+                self.transportServer = SocketServer_Multiplex()
+            else:
+                raise errors.PyroError("invalid server type '%s'" % config.SERVERTYPE)
+            self.transportServer.init(self, host, port, unixsocket)
         #: The location (str of the form ``host:portnumber``) on which the Daemon is listening
         self.locationStr = self.transportServer.locationStr
         log.debug("daemon created on %s - %s (pid %d)", self.locationStr, socketutil.family_str(self.transportServer.sock), os.getpid())
@@ -1278,7 +1356,8 @@ class Daemon(object):
                 _log_wiredata(log, "daemon wiredata received", msg)
             if msg.type == message.MSG_PING:
                 # return same seq, but ignore any data (it's a ping, not an echo). Nothing is deserialized.
-                msg = message.Message(message.MSG_PING, b"pong", msg.serializer_id, 0, msg.seq, annotations=self.__annotations(), hmac_key=self._pyroHmacKey)
+                msg = message.Message(message.MSG_PING, b"pong", msg.serializer_id, 0, msg.seq,
+                                      annotations=self.__annotations(), hmac_key=self._pyroHmacKey)
                 if config.LOGWIRE:
                     _log_wiredata(log, "daemon wiredata sending", msg)
                 conn.send(msg.to_bytes())
@@ -1293,7 +1372,10 @@ class Daemon(object):
                 # normal deserialization of remote call arguments
                 objId, method, vargs, kwargs = serializer.deserializeCall(msg.data, compressed=msg.flags & message.FLAGS_COMPRESSED)
             current_context.client = conn
-            current_context.client_sock_addr = conn.sock.getpeername()   # store this because on oneway calls the socket will be disconnected
+            try:
+                current_context.client_sock_addr = conn.sock.getpeername()   # store, because on oneway calls, socket will be disconnected
+            except socket.error:
+                current_context.client_sock_addr = None   # sometimes getpeername() doesn't work...
             current_context.seq = msg.seq
             current_context.annotations = msg.annotations
             current_context.msg_flags = msg.flags
@@ -1495,7 +1577,8 @@ class Daemon(object):
             flags |= message.FLAGS_COMPRESSED
         annotations = dict(annotations or {})
         annotations.update(self.annotations())
-        msg = message.Message(message.MSG_RESULT, data, serializer.serializer_id, flags, seq, annotations=annotations, hmac_key=self._pyroHmacKey)
+        msg = message.Message(message.MSG_RESULT, data, serializer.serializer_id, flags, seq,
+                              annotations=annotations, hmac_key=self._pyroHmacKey)
         if config.LOGWIRE:
             _log_wiredata(log, "daemon wiredata sending (error response)", msg)
         connection.send(msg.to_bytes())
@@ -1532,7 +1615,10 @@ class Daemon(object):
             # register a custom serializer for the type to automatically return proxies
             # we need to do this for all known serializers
             for ser in util._serializers.values():
-                ser.register_type_replacement(type(obj_or_class), pyroObjectToAutoProxy)
+                if inspect.isclass(obj_or_class):
+                    ser.register_type_replacement(obj_or_class, pyroObjectToAutoProxy)
+                else:
+                    ser.register_type_replacement(type(obj_or_class), pyroObjectToAutoProxy)
         # register the object/class in the mapping
         self.objectsById[obj_or_class._pyroId] = obj_or_class
         return self.uriFor(objectId)
@@ -1674,7 +1760,11 @@ class Daemon(object):
         __lazy_dict_iterator_types = (type({}.keys()), type({}.values()), type({}.items()))
 
     def _streamResponse(self, data, client):
-        if isinstance(data, collections.Iterator) or inspect.isgenerator(data):
+        if sys.version_info < (3, 4):
+            from collections import Iterator
+        else:
+            from collections.abc import Iterator
+        if isinstance(data, Iterator) or inspect.isgenerator(data):
             if config.ITER_STREAMING:
                 if type(data) in self.__lazy_dict_iterator_types:
                     raise errors.PyroError("won't serialize or stream lazy dict iterators, convert to list yourself")
@@ -1773,6 +1863,20 @@ class _CallContext(threading.local):
         self.correlation_id = values["correlation_id"]
         self.client_sock_addr = values["client_sock_addr"]
 
+    def track_resource(self, resource):
+        """keep a weak reference to the resource to be tracked for this connection"""
+        if self.client:
+            self.client.tracked_resources.add(resource)
+        else:
+            raise errors.PyroError("cannot track resource on a connectionless call")
+
+    def untrack_resource(self, resource):
+        """no longer track the resource for this connection"""
+        if self.client:
+            self.client.tracked_resources.discard(resource)
+        else:
+            raise errors.PyroError("cannot untrack resource on a connectionless call")
+
 
 class _OnewayCallThread(threading.Thread):
     def __init__(self, target, args, kwargs):
@@ -1940,3 +2044,18 @@ class SerializedBlob(object):
 
 current_context = _CallContext()
 """the context object for the current call. (thread-local)"""
+
+
+# 'async' keyword backwards compatibility for Python versions older than 3.7. New code should not use this!
+if sys.version_info < (3, 7):
+    def asyncproxy(proxy, asynchronous=True, **kwargs):
+        """convenience method to set proxy to asynchronous or sync mode."""
+        if kwargs:
+            kword = list(kwargs.keys())
+            if kword != ["async"]:
+                raise TypeError("asyncproxy() got an unexpected keyword argument '{:s}'".format(kword[0]))
+            asynchronous = kwargs["async"]
+        proxy._pyroAsync(asynchronous)
+    current_module = sys.modules[__name__]
+    pyro4_module = __import__("Pyro4")
+    current_module.__dict__["async"] = pyro4_module.__dict__["async"] = asyncproxy
