@@ -455,6 +455,8 @@ class PickleSerializer(SerializerBase):
         def copyreg_function(obj):
             return replacement_function(obj).__reduce__()
 
+        if object_type is type or not inspect.isclass(object_type):
+            raise ValueError("refusing to register replacement for a non-type or the type 'type' itself")
         try:
             copyreg.pickle(object_type, copyreg_function)
         except TypeError:
@@ -485,6 +487,8 @@ class CloudpickleSerializer(SerializerBase):
         def copyreg_function(obj):
             return replacement_function(obj).__reduce__()
 
+        if object_type is type or not inspect.isclass(object_type):
+            raise ValueError("refusing to register replacement for a non-type or the type 'type' itself")
         try:
             copyreg.pickle(object_type, copyreg_function)
         except TypeError:
@@ -515,6 +519,8 @@ class DillSerializer(SerializerBase):
         def copyreg_function(obj):
             return replacement_function(obj).__reduce__()
 
+        if object_type is type or not inspect.isclass(object_type):
+            raise ValueError("refusing to register replacement for a non-type or the type 'type' itself")
         try:
             copyreg.pickle(object_type, copyreg_function)
         except TypeError:
@@ -600,6 +606,8 @@ class SerpentSerializer(SerializerBase):
             else:
                 serpent_serializer._serialize(replaced, outputstream, indentlevel)
 
+        if object_type is type or not inspect.isclass(object_type):
+            raise ValueError("refusing to register replacement for a non-type or the type 'type' itself")
         serpent.register_class(object_type, custom_serializer)
 
     @classmethod
@@ -651,6 +659,8 @@ class JsonSerializer(SerializerBase):
 
     @classmethod
     def register_type_replacement(cls, object_type, replacement_function):
+        if object_type is type or not inspect.isclass(object_type):
+            raise ValueError("refusing to register replacement for a non-type or the type 'type' itself")
         cls.__type_replacements[object_type] = replacement_function
 
 
@@ -668,12 +678,12 @@ class MsgpackSerializer(SerializerBase):
 
     def loadsCall(self, data):
         data = self._convertToBytes(data)
-        obj, method, vargs, kwargs = msgpack.unpackb(data, encoding="utf-8", object_hook=self.object_hook)
+        obj, method, vargs, kwargs = msgpack.unpackb(data, raw=False, object_hook=self.object_hook)
         return obj, method, vargs, kwargs
 
     def loads(self, data):
         data = self._convertToBytes(data)
-        return msgpack.unpackb(data, encoding="utf-8", object_hook=self.object_hook, ext_hook=self.ext_hook)
+        return msgpack.unpackb(data, raw=False, object_hook=self.object_hook, ext_hook=self.ext_hook)
 
     def default(self, obj):
         replacer = self.__type_replacements.get(type(obj), None)
@@ -718,6 +728,8 @@ class MsgpackSerializer(SerializerBase):
 
     @classmethod
     def register_type_replacement(cls, object_type, replacement_function):
+        if object_type is type or not inspect.isclass(object_type):
+            raise ValueError("refusing to register replacement for a non-type or the type 'type' itself")
         cls.__type_replacements[object_type] = replacement_function
 
 
@@ -739,6 +751,7 @@ def get_serializer_by_id(sid):
     except KeyError:
         raise errors.SerializeError("no serializer available for id %d" % sid)
 
+
 # determine the serializers that are supported
 try:
     import cPickle as pickle
@@ -784,8 +797,8 @@ try:
     else:
         ver = serpent.__version__
     ver = tuple(map(int, ver.split(".")))
-    if ver < (1, 23):  # serpent 1.23 required
-        raise RuntimeError("requires serpent 1.23 or better")
+    if ver < (1, 24):  # serpent 1.24 required
+        raise RuntimeError("requires serpent 1.24 or better")
     _ser = SerpentSerializer()
     _serializers["serpent"] = _ser
     _serializers_by_id[_ser.serializer_id] = _ser
